"""Access modifiers for Python."""

from functools import wraps
from sys import _getframe as getframe
from typing import Callable, TypeVar


class AccessException(Exception):
    """Exception raised when a private or protected method is called from outside the class."""


ReturnType = TypeVar('ReturnType')


def privatemethod(method: Callable[..., ReturnType]) -> Callable[..., ReturnType]:
    """Decorator that creates a private method."""
    method_class_qualname = getframe(1).f_locals.get("__qualname__")
    @wraps(method)
    def private_method_wrapper(*args, **kwargs) -> ReturnType:
        """Wrap the original method to make it private."""
        caller_frame = getframe(1)
        caller_code = caller_frame.f_code
        caller_name = caller_code.co_name
        while caller_name.startswith("<"):  # Code is a <lambda>, <dictcomp>, <listcomp>, or other non-method code block
            caller_frame = caller_frame.f_back
            caller_code = caller_frame.f_code
            caller_name = caller_code.co_name
        caller_instance = caller_frame.f_locals.get("self")
        if caller_instance is not args[0]:
            raise AccessException(f"Attempted call to private method {method} from another object")
        # Look up the calling method to see if it's defined in the same class as the private method
        for caller_class in caller_instance.__class__.mro():
            caller = caller_class.__dict__.get(caller_name)
            if caller and caller.__code__ == caller_code and method_class_qualname == caller_class.__qualname__:
                return method(*args, **kwargs)
        raise AccessException(f"Attempted call to private method {method} from a sub- or superclass method")
    return private_method_wrapper


def protectedmethod(method: Callable[..., ReturnType]) -> Callable[..., ReturnType]:
    """Decorator that creates a protected method."""
    @wraps(method)
    def protected_method_wrapper(*args, **kwargs) -> ReturnType:
        """Wrap the original method to make it protected."""
        caller_frame = getframe(1)
        caller_instance = caller_frame.f_locals.get("self")
        if caller_instance is not args[0]:
            raise AccessException(f"Attempted call to protected method {method} from another object")
        return method(*args, **kwargs)
    return protected_method_wrapper



"""Unit tests for the protected method access modifier."""

import unittest

from ..access_modifiers import AccessException, protectedmethod, privatemethod


class ProtectedMethodTest(unittest.TestCase):
    """Unit tests for the @protectedmethod decorator."""

    # pylint: disable=missing-docstring

    class Class:
        @protectedmethod
        def protected_method(self):  # pylint: disable=no-self-use
            return "Class.protected_method"

        def public_method(self):
            return "Class.public_method -> " + self.protected_method()

    def test_protected_method(self):
        """Test that accessing a protected method throws an exception."""
        self.assertRaises(AccessException, self.Class().protected_method)
        self.assertRaises(AccessException, self.Class.protected_method, self.Class())

    def test_call_protected_method_via_public_method(self):
        """Test the accessing a protected method via a public method is allowed."""
        self.assertEqual("Class.public_method -> Class.protected_method", self.Class().public_method())

    def test_call_protected_method_via_protected_method(self):
        """Test the accessing a protected method via a another protected method is allowed."""

        class Subclass(self.Class):
            @protectedmethod
            def another_protected_method(self):
                return "Subclass.another_protected_method -> " + self.protected_method()

            def another_public_method(self):
                return "Subclass.another_public_method -> " + self.another_protected_method()

        self.assertEqual(
            "Subclass.another_public_method -> Subclass.another_protected_method -> Class.protected_method",
            Subclass().another_public_method())

    def test_call_protected_method_via_public_method_in_subclass(self):
        """Test that accessing a protected method via a public method in a subclass is allowed."""

        class Subclass(self.Class):
            def public_method(self):
                return "Subclass.public_method -> " + self.protected_method()

        self.assertEqual("Subclass.public_method -> Class.protected_method", Subclass().public_method())

    def test_call_protected_method_via_overridden_public_method(self):
        """Test that accessing a protected method via an overridden public method in a subclass is allowed."""

        class Subclass(self.Class):
            def public_method(self):
                return "Subclass.public_method -> " + super().public_method()

        self.assertEqual(
            "Subclass.public_method -> Class.public_method -> Class.protected_method", Subclass().public_method())

    def test_call_protected_method_via_public_method_using_super(self):
        """Test that accessing a protected method via a public method in a subclass is allowed,
        also when using super()."""

        class Subclass(self.Class):
            def public_method(self):
                return "Subclass.public_method -> " + super().protected_method()

        self.assertEqual("Subclass.public_method -> Class.protected_method", Subclass().public_method())

    def test_call_overridden_protected_method_via_overridden_public_method(self):
        """Test that a protected method can be overridden."""

        class Subclass(self.Class):
            @protectedmethod
            def protected_method(self):
                return "Subclass.protected_method -> " + super().protected_method()

            def public_method(self):
                return "Subclass.public_method -> " + super().public_method()

        self.assertEqual(
            "Subclass.public_method -> Class.public_method -> Subclass.protected_method -> Class.protected_method",
            Subclass().public_method())

    def test_override_protected_method_with_protected_method(self):
        """Test that a protected method can be overridden with a protected method."""

        class Subclass(self.Class):
            @protectedmethod
            def protected_method(self):
                return "Subclass.protected_method -> " + super().protected_method()

        self.assertEqual(
            "Class.public_method -> Subclass.protected_method -> Class.protected_method", Subclass().public_method())

    def test_override_protected_method_with_public_method(self):
        """Test that a protected method can be overridden with a public method."""

        class Subclass(self.Class):
            def protected_method(self):
                return "Subclass.protected_method -> " + super().protected_method()

        self.assertEqual(
            "Class.public_method -> Subclass.protected_method -> Class.protected_method", Subclass().public_method())
        self.assertEqual("Subclass.protected_method -> Class.protected_method", Subclass().protected_method())

    def test_override_protected_method_with_private_method(self):
        """Test that a protected method cannot be overridden with a private method."""

        class Subclass(self.Class):
            @privatemethod
            def protected_method(self):
                return "Subclass.protected_method -> " + super().protected_method()  # pragma: nocover

        self.assertEqual("Class.public_method -> Class.protected_method", self.Class().public_method())
        self.assertRaises(AccessException, Subclass().public_method)
        self.assertRaises(AccessException, Subclass().protected_method)



"""Performance tests for the access modifiers."""

import timeit
import unittest


class PerformanceTest(unittest.TestCase):
    """Performance tests for the access modifiers."""

    def setUp(self):
        self.number = 10_000  # How often to repeat the performance test with timeit.

    def test_create_private_method(self):
        """Test the time it takes to create private methods as compared to methods without access modifier."""
        time_without_modifier = timeit.timeit("""
class C:
    def private_method(self):
        pass
""", setup="from access_modifiers import privatemethod", number=self.number)

        time_with_modifier = timeit.timeit("""
class C:
    @privatemethod
    def private_method(self):
        pass
""", setup="from access_modifiers import privatemethod", number=self.number)
        self.assertLess(time_with_modifier, time_without_modifier * 3)

    def test_create_protected_method(self):
        """Test the time it takes to create protected methods as compared to methods without access modifier."""
        time_without_modifier = timeit.timeit("""
class C:
    def public_method(self):
        pass
""", setup="from access_modifiers import protectedmethod", number=self.number)

        time_with_modifier = timeit.timeit("""
class C:
    @protectedmethod
    def protected_method(self):
        pass
""", setup="from access_modifiers import protectedmethod", number=self.number)
        self.assertLess(time_with_modifier, time_without_modifier * 3)

    def test_call_private_method(self):
        """Test the time it takes to call private methods as compared to methods without access modifier."""
        time_without_modifier = timeit.timeit("c.public_method()", setup="""
from access_modifiers import privatemethod
class C:
    def private_method(self):
        pass
    def public_method(self):
        self.private_method()
c = C()
""", number=self.number)

        time_with_modifier = timeit.timeit("c.public_method()", setup="""
from access_modifiers import privatemethod
class C:
    @privatemethod
    def private_method(self):
        pass
    def public_method(self):
        self.private_method()
c = C()
""", number=self.number)
        self.assertLess(time_with_modifier, time_without_modifier * 15)

    def test_call_protected_method(self):
        """Test the time it takes to call protected methods as compared to methods without access modifier."""
        time_without_modifier = timeit.timeit("c.public_method()", setup="""
from access_modifiers import protectedmethod
class C:
    def protected_method(self):
        pass
    def public_method(self):
        self.protected_method()
c = C()
""", number=self.number)

        time_with_modifier = timeit.timeit("c.public_method()", setup="""
from access_modifiers import protectedmethod
class C:
    @protectedmethod
    def protected_method(self):
        pass
    def public_method(self):
        self.protected_method()
c = C()
""", number=self.number)
        self.assertLess(time_with_modifier, time_without_modifier * 10)



"""Unit tests for the private method access modifier."""

import unittest

from ..access_modifiers import AccessException, privatemethod, protectedmethod


class PrivateMethodTest(unittest.TestCase):
    """Unit tests for the @privatemethod decorator."""

    # pylint: disable=missing-docstring

    class Class:
        @privatemethod
        def private_method(self):  # pylint: disable=no-self-use
            return "Class.private_method"

        def public_method(self):
            return "Class.public_method -> " + self.private_method()

    def test_call_private_method_directly(self):
        """Test the accessing a private method throws an exception."""
        self.assertRaises(AccessException, self.Class().private_method)
        self.assertRaises(AccessException, self.Class.private_method, self.Class())

    def test_call_private_method_via_public_method(self):
        """Test the accessing a private method via a public method is allowed."""
        self.assertEqual("Class.public_method -> Class.private_method", self.Class().public_method())

    def test_call_private_method_via_public_method_from_subclass(self):
        """Test the accessing a private method via an overridden public method is allowed."""

        class Subclass(self.Class):
            def public_method(self):
                return "Subclass.public_method -> " + super().public_method()

        self.assertEqual(
            "Subclass.public_method -> Class.public_method -> Class.private_method", Subclass().public_method())

    def test_call_private_method_via_public_method_in_subclass(self):
        """Test that accessing a private method via a public method in a subclass is not allowed."""

        class Subclass(self.Class):
            def public_method(self):
                self.private_method()

        self.assertRaises(AccessException, Subclass().public_method)

    def test_call_private_method_via_public_method_in_subclass_using_super(self):
        """Test the accessing a private method via a public method in a subclass is not allowed,
        not even when using super()."""

        class Subclass(self.Class):
            def public_method(self):
                super().private_method()

        self.assertRaises(AccessException, Subclass().public_method)

    def test_call_private_method_via_list_comprehension(self):
        """Test that accessing a private method via a list comprehension in a public method works."""

        class Class:
            @privatemethod
            def private_method(self):  # pylint: disable=no-self-use
                return "Class.private_method"

            def public_method(self):
                return ["Class.public_method -> " + self.private_method() for _ in range(1)][0]

        self.assertEqual("Class.public_method -> Class.private_method", Class().public_method())

    def test_call_private_method_via_nested_lambda(self):
        """Test that accessing a private method via a nested lambda in a public method works."""

        class Class:
            @privatemethod
            def private_method(self):  # pylint: disable=no-self-use
                return "Class.private_method"

            def public_method(self):
                # pylint: disable=unnecessary-lambda
                inner_lambda_function = lambda: self.private_method()
                outer_lambda_function = lambda: "Class.public_method -> " + inner_lambda_function()
                return outer_lambda_function()

        self.assertEqual("Class.public_method -> Class.private_method", Class().public_method())

    def test_call_private_method_from_try_except_block(self):
        """Test that accessing a private method from a try/except in a public method works."""

        class Class:
            @privatemethod
            def private_method(self):  # pylint: disable=no-self-use
                return "Class.private_method"

            def public_method(self):
                try:
                    return "Class.public_method -> " + self.private_method()
                except AttributeError:  # pragma: nocover
                    pass

        self.assertEqual("Class.public_method -> Class.private_method", Class().public_method())

    def test_override_private_method(self):
        """Test that an overridden private method can't call its super."""

        class Subclass(self.Class):
            @privatemethod
            def private_method(self):
                super().private_method()  # pragma: nocover

        self.assertRaises(AccessException, Subclass().public_method)

    def test_override_private_method_with_protected_method(self):
        """Test that a private method cannot be overridden with a protected method."""

        class Subclass(self.Class):
            @protectedmethod
            def private_method(self):
                return "Subclass.private_method -> " + super().private_method()

        self.assertEqual("Class.public_method -> Class.private_method", self.Class().public_method())
        self.assertRaises(AccessException, Subclass().public_method)
        self.assertRaises(AccessException, Subclass().private_method)



"""Setup access_modifiers."""

import setuptools


setuptools.setup(
    name="access_modifiers",
    version="0.1.4",
    author="Frank Niessink",
    author_email="frank@niessink.com",
    description="Private and protected access modifiers for Python",
    license="Apache 2.0",
    keywords="access modifier,protected,private,oop",
    url="https://github.com/fniessink/access-modifiers",
    packages=setuptools.find_packages(),
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Software Development"])
